@@ -1,9 +1,11 @@
-// This file is part of Retro. License: GPLv3
+// This file is part of retro-gtk. License: GPL-3.0+.
 
 #ifndef RETRO_CORE_H
 #define RETRO_CORE_H
 
+#include "retro-disk-control-callback.h"
 #include "retro-gtk-internal.h"
+#include "retro-module.h"
 #include "retro-rotation.h"
 
 G_BEGIN_DECLS
@@ -17,7 +19,10 @@ typedef struct {
 typedef struct _RetroCoreEnvironmentInternal RetroCoreEnvironmentInternal;
 
 struct _RetroCoreEnvironmentInternal {
+  RetroModule *module;
+  RetroDiskControlCallback *disk_control_callback;
   gchar **media_uris;
+  RetroSystemInfo *system_info;
   gfloat aspect_ratio;
   gboolean overscan;
   RetroPixelFormat pixel_format;
@@ -27,6 +32,7 @@ struct _RetroCoreEnvironmentInternal {
   RetroKeyboardCallback keyboard_callback;
 };
 
+gchar *retro_core_get_name (RetroCore *self);
 void retro_core_set_system_av_info (RetroCore         *self,
                                     RetroSystemAvInfo *system_av_info);
 void retro_core_set_controller_port_device (RetroCore       *self,
